#ifndef _AMP_CONTROL_H
#define _AMP_CONTROL_H

//////////////////////////////////////
//
// Conditional compilation defines
//
//////////////////////////////////////

// Output Tube Choice, only define one at a time
#define OUTPUT_TUBE_100TH
//#define OUTPUT_TUBE_327A
//#define OUTPUT_TUBE_VT_127A
//#define OUTPUT_TUBE_211

//#define AMP_CONTROL_DEBUG // serial port outputs for debugging
//#define HOURS_METER_INITIALIZE // enables code that resets hours meter to zero in setup()
//#define BIAS_DEFAULTS_INITIALIZE //enables code that sets default bias values in EEPROM in setup()

//////////////////////////////////////
//
// Configurables
//
//////////////////////////////////////
#define SECONDS_TO_INPUT_TUBE_WARMUP 30
#define SECONDS_TO_OUTPUT_TUBE_WARMUP 5
// time the bias ramp will take
const long BIAS_ADJUST_MILLISECONDS = 5000;
// must be larger than the adjust time, lets the bias settle before next adjustment
const long BIAS_TOTAL_MILLISECONDS = 30000;
// minimum display update period in milliseconds, makes display less jittery
const unsigned long MIN_DISPLAY_UPDATE_PERIOD = 200; 

// Autobias information for 6BN11
const long IT_BIAS_VOLTAGE_TARGET = 320;
const long IT_PLATE_NEAR_SATURATION_VOLTAGE = 50;
const long IT_PLATE_NEAR_HV_VOLTAGE = 400;
const long IT_BIAS_MULT = 1;
const long IT_BIAS_DIV = 4;
const long IT_BIAS_FAST_MULT = 1;
const long IT_BIAS_FAST_DIV = 3;
#define IT_BIAS_DAC_MAX 0x00000ECC // 125V
#define IT_BIAS_DAC_MIN 0x000008E1 // 75V
// EEPROM default values for writing during EEPROM 
// initialization (only when compile flag is set)
// .033V/count with HV amp chip
#define EEPROM_IT1_BIAS_VALUE 0x00000BE3 // 100.6V
#define EEPROM_IT2_BIAS_VALUE 0x00000BE3 // 99.6V

// Autobias information for output tubes
#if defined (OUTPUT_TUBE_211)
  const bool PT_BIAS = false; // 211 not under uC control
  const unsigned long HV_FAULT_VALUE = 700; // in V, nominal 615
  const unsigned long PT_FAULT_VALUE = 75; // output tube rating in W
  const long PT_BIAS_CURRENT_TARGET = 110; // in mA
  const long PT_BIAS_POWER_LIMIT = 72; // in W

// power tube info for 100TH
#elif defined (OUTPUT_TUBE_100TH)
  const bool PT_BIAS = true; // 100TH under uC control
  const unsigned long HV_FAULT_VALUE = 825; // in V, nominal 750
  const unsigned long PT_FAULT_VALUE = 100; // output tube rating in W
  const long PT_BIAS_CURRENT_TARGET = 120; // in mA TO DO: set to 125 after testing
  const long PT_BIAS_POWER_LIMIT = 93; // in W

  // Autobias gain information for 100TH
  const long PT_BIAS_MULT = 1;
  const long PT_BIAS_DIV = 1;
  const long PT_BIAS_FAST_MULT = 2;
  const long PT_BIAS_FAST_DIV = 1;

  // Autobias hard limits
  #define PT_BIAS_DAC_MAX 0x000005EB // 50V
  #define PT_BIAS_DAC_MIN 0x00000096 // 5V

  // EEPROM default values for writing during EEPROM 
  // initialization (only when compile flag is set)
  // .033V/count with HV amp chip
  #define EEPROM_PT1_BIAS_VALUE 0x000002EE // 25V
  #define EEPROM_PT2_BIAS_VALUE 0x000002EE // 25V
#endif

//////////////////////////////////////
//
// Fixed Design Values
//
//////////////////////////////////////

// Hours Meter defines
#define DISPLAY_HOURS_METER 0
#define EEPROM_HOURS_METER_DEFAULT_VALUE 0x00000000
#define EEPROM_HOURS_METER_ADDRESS 144

// EEPROM addresses for default value storage
#define EEPROM_PT1_BIAS_ADDRESS 128
#define EEPROM_PT2_BIAS_ADDRESS 132
#define EEPROM_IT1_BIAS_ADDRESS 136
#define EEPROM_IT2_BIAS_ADDRESS 140

// Digital Input Pins
#define POWER_BUTTON 2
#define POWER_BUTTON_PRESSED LOW
#define BIAS_BUTTON 3
#define BIAS_BUTTON_PRESSED LOW

// Analog input pins
#define CHANNEL_A_DRIVER_SENSE_PIN A0
#define CHANNEL_B_DRIVER_SENSE_PIN A1
#define CHANNEL_A_FINAL_SENSE_PIN A2
#define CHANNEL_B_FINAL_SENSE_PIN A3
// #define SILENCE_SENSE_PIN A6
#define HV_SENSE_PIN A7

// Digital Output Pins
#define POWER_SSR 10

// Analog Outputs
#define PT1_DAC_OUTPUT MCP4728_CHANNEL_A // power tube 1
#define PT2_DAC_OUTPUT MCP4728_CHANNEL_B // power tube 2
#define IT1_DAC_OUTPUT MCP4728_CHANNEL_C // input tube 1
#define IT2_DAC_OUTPUT MCP4728_CHANNEL_D // input tube 2

// DACs
#define DAC_ADDRESS 0x64

// LCD
#define LCD_ADDRESS 0

//////////////////////////////////////
//
// Type/function definitions
//
//////////////////////////////////////

// amplifier fault codes
enum amp_fault { FAULT_NONE = 0, FAULT_HV, FAULT_PT1_OVERPOWER, FAULT_PT2_OVERPOWER, FAULT_IT1_BIAS_HIGH, FAULT_IT2_BIAS_HIGH,
                 FAULT_IT1_BIAS_LOW, FAULT_IT2_BIAS_LOW, FAULT_OT1_BIAS_HIGH, FAULT_OT2_BIAS_HIGH, FAULT_OT1_BIAS_LOW,
                 FAULT_OT2_BIAS_LOW, FAULT_BIAS_ARITHMETIC_PT, FAULT_BIAS_ARITHMETIC_IT, FAULT_BIAS_ADJUST};

// amplifier state variable type
enum amp_state { AMPLIFIER_OFF, AMPLIFIER_WARMUP, AMPLIFIER_ON };

// bias process state variable type
enum bias_state_type { BIAS_IDLE = 0, BIAS_ADJUST };

// main state task functions
void amplifier_off_task();
void amplifier_warmup_task();
void amplifier_on_task();

// state transition functions
void setup();
void transition_to_amplifier_off(amp_fault fault_code);

// state utility functions
void bias_subtask();
void read_ADCs_and_update_LCD(unsigned long amp_start_time);
int check_for_faults();
void update_builtinLED();

// minor utility functions
void print_by_digit(unsigned int print_value, int num_digits);
unsigned int decimal_place_divisor (int num_digits);

#endif

// Reference Calculations and Measurements

// DAC
// 12 bit DAC, 2.048V reference
// .0005V per count
//
// HV Amp
// Gain = 66.7
// .03335V per count
//
// .03335V per step
// 136.6V Max out (134)
// 2.048V max input at chip with 1X ref gain
// 144V HV PS to chip
// Vg 39.45V 1182 0x049E
// Vs 101.5V 3043 0x0BE3

// High Side Current Sense Board
//
// 10 bit ADC, 0.55V ref
// A = 4.1, offset 1.25V
// 1.0490417 Ohm Sense
// 480 cnts for 120mA
// 120mA -> .125885V -> .516129V -> .258065V

// Gain Measurements IT
//
// 102.1V -> 240V
// 103.1V -> 195V
// A = ~-50

// Gain Measurements OT
//
// 36.47V -> 117.5mA
// 34.47V -> 110.5mA
// A = ~3.5mA/V