//////////////////////////////////////
//
// Corona Tube Amp Control
//
// Heath Workman Sept 2022
// 
// Feel free to use this in your own projects
// If you use it in a commercial product, I'd 
// appreciate it if you give me credit somewhere
//
//////////////////////////////////////
//
// Inputs:
//         Power button
//         Bias Button
//         ADC HVPS
//         ADC Silence Sense (feature not implemented)
//         2 ADC Output Stage current sense
//         2 ADC Input Stage voltage sense
//         Total: 6 ADC 2 digital
//
// Outputs: 
//          SSR for main power
//          2 DACS for output stage bias control
//          2 DACS for input stage bias control
//          Built-in LED - Heartbeat
//          Display
//          Total: 4 DAC, 1 Display, 1 Digital
//
//////////////////////////////////////

#include <EEPROM.h>
#include <Wire.h>
#include <Adafruit_MCP4728.h>
#include <Adafruit_LiquidCrystal.h>
#include "AmpControl.h"

Adafruit_MCP4728 mcp; // 12 bit 4 channel DAC
uint32_t PT1_dac_val;
uint32_t PT2_dac_val;
uint32_t IT1_dac_val;
uint32_t IT2_dac_val;

Adafruit_LiquidCrystal lcd(LCD_ADDRESS);

amp_state amplifier_state;

//Global amplifier fault code
amp_fault amplifier_fault_code = FAULT_NONE;

// ADC value global variables, 10 bit
// 1V = 1 count 10M->5379R divider
unsigned int channel_A_driver_sense_value; 
unsigned int channel_B_driver_sense_value;
// 1mA = 2 counts 1.0742R sense resistor or 4 counts with High-Side sense
unsigned long channel_A_final_sense_value; 
unsigned long channel_B_final_sense_value;
// unsigned int silence_sense_value;
// 1V = 1 count 10M->5379R divider
unsigned long HV_sense_value; 

// Calculated from ADC values
unsigned long channel_A_final_power = 0;
unsigned long channel_B_final_power = 0;

unsigned long hours_meter;
unsigned long start_time;

// main function, runs in a loop
void loop() 
{
  switch (amplifier_state)
  {
    case AMPLIFIER_OFF : 
      amplifier_off_task();
      break;
    case AMPLIFIER_WARMUP :
      amplifier_warmup_task();
      break;
    case AMPLIFIER_ON :
      amplifier_on_task();
      break;
    default :
      break;
  }
}

void amplifier_off_task() 
{
  update_builtinLED(); 
  if (digitalRead(POWER_BUTTON) == POWER_BUTTON_PRESSED)
  {
    amplifier_state = AMPLIFIER_WARMUP;
  }
  amplifier_fault_code = FAULT_NONE;  
}

void amplifier_warmup_task() 
{
  // retrieve DAC default values from EEPROM
  EEPROM.get(EEPROM_PT1_BIAS_ADDRESS, PT1_dac_val);
  EEPROM.get(EEPROM_PT2_BIAS_ADDRESS, PT2_dac_val);
  EEPROM.get(EEPROM_IT1_BIAS_ADDRESS, IT1_dac_val);
  EEPROM.get(EEPROM_IT2_BIAS_ADDRESS, IT2_dac_val);
  EEPROM.get(EEPROM_HOURS_METER_ADDRESS, hours_meter);
  
  // set initial DAC output values (2.048V internal reference X1)
  mcp.setChannelValue(PT1_DAC_OUTPUT, PT1_dac_val, MCP4728_VREF_INTERNAL, MCP4728_GAIN_1X);
  mcp.setChannelValue(PT2_DAC_OUTPUT, PT2_dac_val, MCP4728_VREF_INTERNAL, MCP4728_GAIN_1X);
  mcp.setChannelValue(IT1_DAC_OUTPUT, IT1_dac_val, MCP4728_VREF_INTERNAL, MCP4728_GAIN_1X);
  mcp.setChannelValue(IT2_DAC_OUTPUT, IT2_dac_val, MCP4728_VREF_INTERNAL, MCP4728_GAIN_1X);

  // set up display for run mode
  lcd.setBacklight(HIGH);
  lcd.setCursor(0, 0);
  lcd.print("V1A:    V V1B:    V ");
  lcd.setCursor(0, 1);
  lcd.print("V2:     W         mA");
  lcd.setCursor(0, 2);
  lcd.print("V3:     W         mA");
  lcd.setCursor(0, 3);
  lcd.print("HV:     V Warmup:   ");

  // turn on SSR
  digitalWrite(POWER_SSR, HIGH);
  
  // mark start time
  start_time = millis();

  //#ifdef AMP_CONTROL_DEBUG
  //Serial.println("PT DAC vals");
  //Serial.println(PT1_dac_val);
  //Serial.println(PT2_dac_val); 
  //Serial.println("start_time, hours_meter");
  //Serial.println(start_time);
  //Serial.println(hours_meter);
  //#endif
  
  //while tubes are warming up
  while((unsigned long)(millis() - start_time) <= (SECONDS_TO_OUTPUT_TUBE_WARMUP * 1000) || 
        (unsigned long)(millis() - start_time) <= (SECONDS_TO_INPUT_TUBE_WARMUP * 1000))
  {
    read_ADCs_and_update_LCD(start_time);
    update_builtinLED();
    bias_subtask();
    if(check_for_faults())
    {
      bias_subtask(); // allow bias subtask to perform houskeeping from fault
      transition_to_amplifier_off(amplifier_fault_code);
      break;    
    }
  }
  // transition to run state if no faults occurred
  if(amplifier_state == AMPLIFIER_WARMUP)
  {
    amplifier_state = AMPLIFIER_ON;
    lcd.setCursor(10, 3);
    lcd.print("HRS: ");
  }
}

void amplifier_on_task() 
{ 
  update_builtinLED(); 
  read_ADCs_and_update_LCD(DISPLAY_HOURS_METER);
  if(check_for_faults())
    transition_to_amplifier_off(amplifier_fault_code);
  if(digitalRead(POWER_BUTTON) == POWER_BUTTON_PRESSED)
    transition_to_amplifier_off(FAULT_NONE);
  bias_subtask();
}

void setup() 
{
  // initialize SSR to 'OFF'
  pinMode(POWER_SSR, OUTPUT);
  digitalWrite(POWER_SSR, LOW);
  
  // EEPROM bias initialization code
  #ifdef BIAS_DEFAULTS_INITIALIZE
    EEPROM.put(EEPROM_PT1_BIAS_ADDRESS, uint32_t(EEPROM_PT1_BIAS_VALUE));
    EEPROM.put(EEPROM_PT2_BIAS_ADDRESS, uint32_t(EEPROM_PT2_BIAS_VALUE));
    EEPROM.put(EEPROM_IT1_BIAS_ADDRESS, uint32_t(EEPROM_IT1_BIAS_VALUE));
    EEPROM.put(EEPROM_IT2_BIAS_ADDRESS, uint32_t(EEPROM_IT2_BIAS_VALUE));
  #endif
  
  #ifdef HOURS_METER_INITIALIZE
    EEPROM.put(EEPROM_HOURS_METER_ADDRESS, uint32_t(EEPROM_HOURS_METER_DEFAULT_VALUE));
  #endif
  
  // get DAC output default values
  mcp.begin(DAC_ADDRESS);
  EEPROM.get(EEPROM_PT1_BIAS_ADDRESS, PT1_dac_val);
  EEPROM.get(EEPROM_PT2_BIAS_ADDRESS, PT2_dac_val);
  EEPROM.get(EEPROM_IT1_BIAS_ADDRESS, IT1_dac_val);
  EEPROM.get(EEPROM_IT2_BIAS_ADDRESS, IT2_dac_val);

  // set initial DAC output values (2.048V internal reference X1)
  mcp.setChannelValue(PT1_DAC_OUTPUT, PT2_dac_val, MCP4728_VREF_INTERNAL, MCP4728_GAIN_1X);
  mcp.setChannelValue(PT2_DAC_OUTPUT, PT2_dac_val, MCP4728_VREF_INTERNAL, MCP4728_GAIN_1X);
  mcp.setChannelValue(IT1_DAC_OUTPUT, IT1_dac_val, MCP4728_VREF_INTERNAL, MCP4728_GAIN_1X);
  mcp.setChannelValue(IT2_DAC_OUTPUT, IT2_dac_val, MCP4728_VREF_INTERNAL, MCP4728_GAIN_1X); 
  
  #ifdef AMP_CONTROL_DEBUG
  Serial.begin(115200);
  while (!Serial)
    delay(10);
  Serial.println("Debug mode enabled.");
  //Serial.println("DAC_vals, PT1 PT2 IT1 IT2");
  //Serial.println(PT1_dac_val);
  //Serial.println(PT2_dac_val);
  //Serial.println(IT1_dac_val);
  //Serial.println(IT2_dac_val);  
  #endif

  //set up digital inputs
  pinMode(POWER_BUTTON, INPUT_PULLUP);
  pinMode(BIAS_BUTTON, INPUT_PULLUP);
  
  // set up LEDs
  pinMode(LED_BUILTIN, OUTPUT);
  digitalWrite(LED_BUILTIN, HIGH);

  // initialize analog reference
  analogReference(INTERNAL0V55);

  // initialize 20x4 LCD
  lcd.begin(20, 4);
  lcd.setBacklight(LOW);

  // initialize amplifer state
  amplifier_state = AMPLIFIER_OFF;
}

void transition_to_amplifier_off(amp_fault fault_code) 
{
  unsigned long seconds_on = (unsigned long)(millis() - start_time) / 1000;
  
  amplifier_state = AMPLIFIER_OFF;
  bias_subtask(); // call this so that bias subtask can go to idle state
  digitalWrite(POWER_SSR, LOW);
  lcd.clear();
  lcd.setCursor(0, 0);

  if(fault_code == FAULT_NONE)
    lcd.print(" OFF");
  else if(fault_code == FAULT_HV)
    lcd.print(" HV PS Over-voltage");
  else if(fault_code == FAULT_PT1_OVERPOWER)
    lcd.print(" PT1 Over-power");
  else if(fault_code == FAULT_PT2_OVERPOWER)
    lcd.print(" PT2 Over-power");
  else if(fault_code == FAULT_IT1_BIAS_HIGH)
    lcd.print(" IT1 Bias High");
  else if(fault_code == FAULT_IT2_BIAS_HIGH)
    lcd.print(" IT2 Bias High");
  else if(fault_code == FAULT_IT1_BIAS_LOW)
    lcd.print(" IT1 Bias Low");
  else if(fault_code == FAULT_IT2_BIAS_LOW)
    lcd.print(" IT2 Bias Low");
  else if(fault_code == FAULT_OT1_BIAS_HIGH)
    lcd.print(" OT1 Bias High");
  else if(fault_code == FAULT_OT2_BIAS_HIGH)
    lcd.print(" OT2 Bias High");
  else if(fault_code == FAULT_OT1_BIAS_LOW)
    lcd.print(" OT1 Bias Low");
  else if(fault_code == FAULT_OT2_BIAS_LOW)
    lcd.print(" OT2 Bias Low");
  else if(fault_code == FAULT_BIAS_ARITHMETIC_PT)
    lcd.print(" PT Bias Math Error");
  else if(fault_code == FAULT_BIAS_ARITHMETIC_IT)
    lcd.print(" IT Bias Math Error");
  else if(fault_code == FAULT_BIAS_ADJUST)
    lcd.print(" Bias Adjust Error");
  lcd.setBacklight(LOW);
 
  //#ifdef AMP_CONTROL_DEBUG
  //Serial.println("seconds_on, hours_meter before");
  //Serial.println(seconds_on);
  //Serial.println(hours_meter);
  //#endif

  hours_meter += seconds_on;

  //update hours meter
  EEPROM.put(EEPROM_HOURS_METER_ADDRESS, hours_meter);

  // store last good bias values to EEPROM
  if(!amplifier_fault_code)
  {
    EEPROM.put(EEPROM_PT1_BIAS_ADDRESS, PT1_dac_val);
  //  EEPROM.put(EEPROM_PT2_BIAS_ADDRESS, PT2_dac_val);
    EEPROM.put(EEPROM_IT1_BIAS_ADDRESS, IT1_dac_val);
  //  EEPROM.put(EEPROM_IT2_BIAS_ADDRESS, IT2_dac_val);
  }
  else if(amplifier_fault_code == FAULT_PT1_OVERPOWER)
    EEPROM.put(EEPROM_PT1_BIAS_ADDRESS, PT1_dac_val >> 1);
  //else if(amplifier_fault_code == FAULT_PT2_OVERPOWER)
  //  EEPROM.put(EEPROM_PT2_BIAS_ADDRESS, PT2_dac_val >> 1);

  //#ifdef AMP_CONTROL_DEBUG
  //Serial.println("hours_meter after");
  //Serial.println(hours_meter);
  //#endif

  // give plenty of time for button press to stop
  delay(10000);
}

void bias_subtask()
{
  static bias_state_type bias_state = BIAS_IDLE;
  static unsigned long bias_start_time = 0;
  
  static long pt_bias_power_current_target = 0;

  static long DAC_output_change_PT1 = 0; //number of counts that bias output needs to move
  static long change_interval_PT1 = 0; // number of milliseconds between count increases
  static long change_count_PT1 = 0; // number of counts that we have already incremented
  static long new_DAC_value_PT1 = 0; // calculated new DAC value for error checking
  static long channel_A_final_plate_current = 0;
  static bool PT1_bias_complete = false;
  
  static long DAC_output_change_IT1 = 0; //number of counts that bias output needs to move
  static long change_interval_IT1 = 0; // number of milliseconds between count increases
  static long change_count_IT1 = 0; // number of counts that we have already incremented
  static long new_DAC_value_IT1 = 0; // calculated new DAC value for error checking
  static bool IT1_bias_complete = false;

  switch (bias_state)
  {
    case BIAS_IDLE :
      if(digitalRead(BIAS_BUTTON) == BIAS_BUTTON_PRESSED)
      {
        //initialize bias process
        bias_start_time = millis();
        pt_bias_power_current_target = 0;
        DAC_output_change_PT1 = 0;
        change_interval_PT1 = 0;
        change_count_PT1 = 0;
        channel_A_final_plate_current = 0;
        PT1_bias_complete = false;
        DAC_output_change_IT1 = 0;
        change_interval_IT1 = 0;
        change_count_IT1 = 0;
        IT1_bias_complete = false;

        // power tubes
        if(PT_BIAS && (unsigned long)(millis() - start_time) >= (SECONDS_TO_OUTPUT_TUBE_WARMUP * 1000))
	      {
          // subtract out feedback resistor current
          channel_A_final_plate_current = channel_A_final_sense_value > 14 ? (channel_A_final_sense_value - 14) : 0;
          // if HV is too high to set PT at ideal current, dial current target back
          if((((long)HV_sense_value * PT_BIAS_CURRENT_TARGET) / (long)1000) > PT_BIAS_POWER_LIMIT)
            pt_bias_power_current_target = ((PT_BIAS_POWER_LIMIT * (long)1000) / (long)HV_sense_value); // set power limit to 75 for testing
          else
            pt_bias_power_current_target = PT_BIAS_CURRENT_TARGET;
          // if power tube is above power target adjust faster
		      if(channel_A_final_power > PT_BIAS_POWER_LIMIT)
            DAC_output_change_PT1 = ((((pt_bias_power_current_target << 2) - channel_A_final_plate_current) * PT_BIAS_FAST_MULT) / PT_BIAS_FAST_DIV);
		      else
            DAC_output_change_PT1 = ((((pt_bias_power_current_target << 2) - channel_A_final_plate_current) * PT_BIAS_MULT) / PT_BIAS_DIV);
          new_DAC_value_PT1 = (long)(PT1_dac_val + DAC_output_change_PT1);
          // error check
          if(new_DAC_value_PT1 & 0xFFFFF000) // if we set bits we shouldn't have
            amplifier_fault_code = FAULT_BIAS_ARITHMETIC_PT;
          else if(new_DAC_value_PT1 > PT_BIAS_DAC_MAX)
            amplifier_fault_code = FAULT_OT1_BIAS_HIGH;
          else if(new_DAC_value_PT1 < PT_BIAS_DAC_MIN)
            amplifier_fault_code = FAULT_OT1_BIAS_LOW;
          change_interval_PT1 = BIAS_ADJUST_MILLISECONDS / DAC_output_change_PT1;
          if (change_interval_PT1 < 0)
            change_interval_PT1 *= -1;
          #ifdef AMP_CONTROL_DEBUG
          Serial.println("PT1_dac_val");
          Serial.println(PT1_dac_val);
          Serial.println("channel_A_final_sense_value");
          Serial.println(channel_A_final_sense_value);
          Serial.println("channel_A_final_plate_current");
          Serial.println(channel_A_final_plate_current);
          Serial.println("pt_bias_power_current_target");
          Serial.println(pt_bias_power_current_target);
          Serial.println("DAC_output_change_PT1");
          Serial.println(DAC_output_change_PT1);
          Serial.println("new_DAC_value_PT1");
          Serial.println(new_DAC_value_PT1);
          Serial.println("change_interval_PT1");
          Serial.println(change_interval_PT1);
          #endif
	      }

        // input tubes
        if((unsigned long)(millis() - start_time) >= (SECONDS_TO_INPUT_TUBE_WARMUP * 1000))
        {
          // if driver plate is within 15 volts of the supply voltage or near saturation, adjust faster
          if((long)channel_A_driver_sense_value > IT_PLATE_NEAR_HV_VOLTAGE || (long)channel_A_driver_sense_value < IT_PLATE_NEAR_SATURATION_VOLTAGE)
            DAC_output_change_IT1 = ((((long)channel_A_driver_sense_value - IT_BIAS_VOLTAGE_TARGET) * IT_BIAS_FAST_MULT) / IT_BIAS_FAST_DIV);
          else
          {
            DAC_output_change_IT1 = ((((long)channel_A_driver_sense_value - IT_BIAS_VOLTAGE_TARGET) * IT_BIAS_MULT) / IT_BIAS_DIV);
            // Give the gain a little bump if bias is close so it will reach the value
            if(DAC_output_change_IT1 == 0 && ((long)channel_A_driver_sense_value - IT_BIAS_VOLTAGE_TARGET) > 0)
              DAC_output_change_IT1 = 1;
            else if(DAC_output_change_IT1 == 0 && ((long)channel_A_driver_sense_value - IT_BIAS_VOLTAGE_TARGET) < 0)
              DAC_output_change_IT1 = -1;
          }
          new_DAC_value_IT1 = (long)(IT1_dac_val + DAC_output_change_IT1);
          // error check
          if(new_DAC_value_IT1 & 0xFFFFF000) // if we set bits we shouldn't have
            amplifier_fault_code = FAULT_BIAS_ARITHMETIC_IT;
          else if(new_DAC_value_IT1 > IT_BIAS_DAC_MAX)
            amplifier_fault_code = FAULT_IT1_BIAS_HIGH;
          else if(new_DAC_value_IT1 < IT_BIAS_DAC_MIN)
            amplifier_fault_code = FAULT_IT1_BIAS_LOW;
          change_interval_IT1 = BIAS_ADJUST_MILLISECONDS / DAC_output_change_IT1;
          if (change_interval_IT1 < 0)
            change_interval_IT1 *= -1;
          #ifdef AMP_CONTROL_DEBUG
          Serial.println("IT1_dac_val");
          Serial.println(IT1_dac_val);
          Serial.println("channel_A_driver_sense_value");
          Serial.println(channel_A_driver_sense_value);
          Serial.println("DAC_output_change_IT1");
          Serial.println(DAC_output_change_IT1);
          Serial.println("new_DAC_value_IT1");
          Serial.println(new_DAC_value_IT1);
          Serial.println("change_interval_IT1");
          Serial.println(change_interval_IT1);
          #endif
        }
        // if we should not transition to adjust state, stay in idle state
        if(amplifier_fault_code || !amplifier_state || (((unsigned long)(millis() - start_time) < (SECONDS_TO_INPUT_TUBE_WARMUP * 1000)) 
           && (!PT_BIAS || (unsigned long)(millis() - start_time) < (SECONDS_TO_OUTPUT_TUBE_WARMUP * 1000))))
        {
          bias_state = BIAS_IDLE;
          #ifdef AMP_CONTROL_DEBUG
          Serial.println("bs - going to idle from idle");
          #endif
        }
        // else transition to adjust state
        else
          bias_state = BIAS_ADJUST;
      }
      break;
    case BIAS_ADJUST : // make sure that cold tubes aren't being adjusted or amplifier isn't going to idle state
      if(amplifier_fault_code || !amplifier_state)
      {
        bias_state = BIAS_IDLE;
        #ifdef AMP_CONTROL_DEBUG
        Serial.println("bs - going to idle early from adjust");
        #endif
      }
      else
      {
        // if time is up, we are done
        if((unsigned long)(millis() - bias_start_time) > BIAS_TOTAL_MILLISECONDS)
          {
            #ifdef AMP_CONTROL_DEBUG
            Serial.println("PT1_dac_val");
            Serial.println(PT1_dac_val);
            Serial.println("IT1_dac_val");
            Serial.println(IT1_dac_val);
            #endif
            bias_state = BIAS_IDLE;
            #ifdef AMP_CONTROL_DEBUG
            Serial.println("bs - going to idle on time from adjust");
            #endif
          }
        // power tube process
        if(DAC_output_change_PT1)
        {
          // if it is time to change an output
		      if((unsigned long)(millis() - bias_start_time) / change_interval_PT1 > change_count_PT1 && !PT1_bias_complete)
          {
            //increment or decrement output variable
            if(DAC_output_change_PT1 > 0)
            {
              PT1_dac_val += (uint32_t)(((unsigned long)(millis() - bias_start_time) / change_interval_PT1) - change_count_PT1);
              if(PT1_dac_val >= new_DAC_value_PT1)
              {
                PT1_bias_complete = true;
                PT1_dac_val = new_DAC_value_PT1;
              }
            }
            else if(DAC_output_change_PT1 < 0)
            {
              PT1_dac_val -= (uint32_t)(((unsigned long)(millis() - bias_start_time) / change_interval_PT1) - change_count_PT1);
              if(PT1_dac_val <= new_DAC_value_PT1)
              {
                PT1_bias_complete = true;
                PT1_dac_val = new_DAC_value_PT1;
              }
            }
            change_count_PT1 += (long)(((unsigned long)(millis() - bias_start_time) / change_interval_PT1) - change_count_PT1);
            // check for fault conditions
            if(PT1_dac_val & 0xFFFFF000) // if we set bits we shouldn't have
              amplifier_fault_code = FAULT_BIAS_ADJUST;
            else if(PT1_dac_val > PT_BIAS_DAC_MAX)
              amplifier_fault_code = FAULT_OT1_BIAS_HIGH;
            else if(PT1_dac_val < PT_BIAS_DAC_MIN)
              amplifier_fault_code = FAULT_OT1_BIAS_LOW;
            // set channel output
            mcp.setChannelValue(PT1_DAC_OUTPUT, PT1_dac_val, MCP4728_VREF_INTERNAL, MCP4728_GAIN_1X);
          } 
	      }
        // input tube process
        if(DAC_output_change_IT1)
        {
          // if it is time to change an output
          if((unsigned long)(millis() - bias_start_time) / change_interval_IT1 > change_count_IT1 && !IT1_bias_complete)
          {
            //increment or decrement output variable
            if(DAC_output_change_IT1 > 0)
            {
              IT1_dac_val += (uint32_t)(((unsigned long)(millis() - bias_start_time) / change_interval_IT1) - change_count_IT1);
              if(IT1_dac_val >= new_DAC_value_IT1)
              {
                IT1_bias_complete = true;
                IT1_dac_val = new_DAC_value_IT1;
              }
            }
            else if(DAC_output_change_IT1 < 0)
            {
              IT1_dac_val -= (uint32_t)(((unsigned long)(millis() - bias_start_time) / change_interval_IT1) - change_count_IT1);
              if(IT1_dac_val <= new_DAC_value_IT1)
              {
                IT1_bias_complete = true;
                IT1_dac_val = new_DAC_value_IT1;
              }
            }
            change_count_IT1 += (long)(((unsigned long)(millis() - bias_start_time) / change_interval_IT1) - change_count_IT1);
            // check for fault conditions
            if(IT1_dac_val & 0xFFFFF000) // if we set bits we shouldn't have
              amplifier_fault_code = FAULT_BIAS_ADJUST;
            else if(IT1_dac_val > IT_BIAS_DAC_MAX)
              amplifier_fault_code = FAULT_IT1_BIAS_HIGH;
            else if(IT1_dac_val < IT_BIAS_DAC_MIN)
              amplifier_fault_code = FAULT_IT1_BIAS_LOW;
            // set channel output
            mcp.setChannelValue(IT1_DAC_OUTPUT, IT1_dac_val, MCP4728_VREF_INTERNAL, MCP4728_GAIN_1X);
          }
        }
      }
      break;
    default :
      bias_state = BIAS_IDLE;
      break;
  }  
}

void read_ADCs_and_update_LCD(unsigned long amp_start_time) 
{
  const unsigned int warmup_time = SECONDS_TO_INPUT_TUBE_WARMUP > SECONDS_TO_OUTPUT_TUBE_WARMUP ? 
                                   SECONDS_TO_INPUT_TUBE_WARMUP : SECONDS_TO_OUTPUT_TUBE_WARMUP;
  
  unsigned long channel_A_final_plate_current;
  unsigned long channel_B_final_plate_current;
  unsigned long hours_display_value;
  unsigned long warmup_display_value;
  bool update_display = false;

  static unsigned long last_update_time = 0;

  static unsigned int previous_channel_A_driver_sense_value = 0;
  static unsigned int previous_channel_B_driver_sense_value = 0;
  static unsigned long previous_channel_A_final_power = 0;
  static unsigned long previous_channel_B_final_power = 0;
  static unsigned long previous_channel_A_final_plate_current = 0;
  static unsigned long previous_channel_B_final_plate_current = 0;
  static unsigned long previous_HV_sense_value = 0;
  static unsigned long previous_hours_display_value = 0xFFFFFFFF;
  static unsigned long previous_warmup_display_value = 0;

  // determine if it is time to update the display
  if((unsigned long)(millis() - last_update_time) > MIN_DISPLAY_UPDATE_PERIOD)
  {
    update_display = true;
    last_update_time = millis();
  }

  // update sense values
  channel_A_driver_sense_value = analogRead(CHANNEL_A_DRIVER_SENSE_PIN);
  channel_B_driver_sense_value = analogRead(CHANNEL_B_DRIVER_SENSE_PIN);
  channel_A_final_sense_value = analogRead(CHANNEL_A_FINAL_SENSE_PIN);
  channel_B_final_sense_value = analogRead(CHANNEL_B_FINAL_SENSE_PIN);
  channel_A_final_plate_current = channel_A_final_sense_value > 14 ? (channel_A_final_sense_value - 14) : 0;
  channel_B_final_plate_current = channel_B_final_sense_value > 14 ? (channel_B_final_sense_value - 14) : 0;
  // silence_sense_value = analogRead(SILENCE_SENSE_PIN);
  HV_sense_value = analogRead(HV_SENSE_PIN);
  // Subtract out feedback resistor current, then calculatue plate power
  channel_A_final_power = (channel_A_final_plate_current * HV_sense_value) / 4000;
  channel_B_final_power = (channel_B_final_plate_current * HV_sense_value) / 4000;
  hours_display_value = hours_meter / 3600;
  warmup_display_value = warmup_time - ((unsigned long)(millis() - amp_start_time) / 1000);

  // update display for each value that has changed
  if(update_display)
  {
    if(previous_channel_A_driver_sense_value != channel_A_driver_sense_value)
    {
      lcd.setCursor(4, 0);
      print_by_digit(channel_A_driver_sense_value, 4);
      previous_channel_A_driver_sense_value = channel_A_driver_sense_value;
    }
    if(previous_channel_B_driver_sense_value != channel_B_driver_sense_value)
    {
      //lcd.setCursor(14, 0);
      //print_by_digit(channel_B_driver_sense_value, 4);
      previous_channel_B_driver_sense_value = channel_B_driver_sense_value;
    }
    if(previous_channel_A_final_power != channel_A_final_power)
    {
      lcd.setCursor(4, 1);
      print_by_digit(channel_A_final_power, 4);
      previous_channel_A_final_power = channel_A_final_power;
    }
    if(((previous_channel_A_final_plate_current + 2) >> 2) != ((channel_A_final_plate_current + 2) >> 2))
    {
      lcd.setCursor(14, 1);
      print_by_digit(((channel_A_final_plate_current + 2) >> 2), 4);
      previous_channel_A_final_plate_current = channel_A_final_plate_current;
    }
    if(previous_channel_B_final_power != channel_B_final_power)
    {
      //lcd.setCursor(4, 2);
      //print_by_digit(channel_B_final_power, 4);
      previous_channel_B_final_power = channel_B_final_power;
    }
    if(((previous_channel_B_final_plate_current + 2) >> 2) != ((channel_B_final_plate_current + 2) >> 2))
    {  
      //lcd.setCursor(14, 2);
      //print_by_digit(((channel_B_final_plate_current + 2) >> 2), 4);
      previous_channel_B_final_plate_current = channel_B_final_plate_current;
    }
    if(previous_HV_sense_value != HV_sense_value)
    {
      lcd.setCursor(4, 3);
      print_by_digit(HV_sense_value, 4);
      previous_HV_sense_value = HV_sense_value;
    }
    if (amp_start_time) // display warmup countdown
    {
      if(previous_warmup_display_value != warmup_display_value)
      {
        lcd.setCursor(17, 3);
        print_by_digit(warmup_display_value, 3);
        previous_warmup_display_value = warmup_display_value;
        previous_hours_display_value = 0xFFFFFFFF;
      }
    }
    else // display hours meter
    {
      if(previous_hours_display_value != hours_display_value)
      {
        lcd.setCursor(14, 3);
        print_by_digit(hours_display_value, 6);
        previous_hours_display_value = hours_display_value;
      }
    }
    update_display = false;
  }
}

int check_for_faults()
{ 
  if(amplifier_fault_code)
    return amplifier_fault_code;
  else
  {
    if(HV_sense_value > HV_FAULT_VALUE)
    {
      amplifier_fault_code = FAULT_HV;
      return FAULT_HV;
    }
    else if(channel_A_final_power > PT_FAULT_VALUE)
    {
      amplifier_fault_code = FAULT_PT1_OVERPOWER;
      return FAULT_PT1_OVERPOWER;
    }
    //else if(channel_B_final_power > PT_FAULT_VALUE)
    //{
    //  amplifier_fault_code = FAULT_PT2_OVERPOWER;
    //  return FAULT_PT2_OVERPOWER;
    //}
    else
      return FAULT_NONE;
  }
}
    
void update_builtinLED() 
{
  // toggle led every 512mSec
  if(millis() & 0x00000200)
    digitalWrite(LED_BUILTIN, LOW);
  else
    digitalWrite(LED_BUILTIN, HIGH);
}

// print one digit at a time of a number
//
// print_value is the value to be printed
// num_digits is the number of digits it is supposted to take up on the display

void print_by_digit(unsigned int print_value, int num_digits)
{
  bool leading_zeros = true;

  for(;num_digits > 0;num_digits--)
  {
    // if digit is nonzero or not leading zero
    if (print_value / decimal_place_divisor(num_digits) || leading_zeros == false)
    {
      lcd.print(print_value / decimal_place_divisor(num_digits));
      leading_zeros = false;
    }
    // else if it is a final digit and is zero
    else if(num_digits == 1)
    {
      lcd.print(print_value / decimal_place_divisor(num_digits));
    }
    // else it is a leading zero
    else
      lcd.print(" ");
    print_value %= decimal_place_divisor(num_digits);
  }
}

// calculate the divisor to divide out lower order digits (truncate value)
//
// num_digits is the number of decimal digits in the value
unsigned int decimal_place_divisor (int num_digits)
{
  unsigned int rv = 1;
  num_digits--;
  for(; num_digits > 0; num_digits--)
  {
    rv *= 10;
  }
  return rv;
}